from votepy.rules.chamberlin_courant_rule import chamberlin_courant
from votepy.rules.greedy_monroe_rule import greedy_monroe
from votepy.rules.k_borda_rule import k_borda
from votepy.rules.sntv_rule import sntv
from votepy.rules.bloc_rule import bloc
import votepy.rules.owa_rule as owa
